import requests

from chaiverse.login_cli import auto_authenticate
from chaiverse.config import BASE_SUBMITTER_URL, BASE_FEEDBACK_URL
from chaiverse.utils import get_url


class _ChaiverseHTTPClient():
    def __init__(self, developer_key=None, hostname=None):
        self.developer_key = developer_key
        self.hostname = hostname

    @property
    def headers(self):
        return {"Authorization": f"Bearer {self.developer_key}"}

    def get(self, endpoint, submission_id=None, **kwargs):
        url = get_url(endpoint, hostname=self.hostname, submission_id=submission_id)
        response = self._request(requests.get, url=url, **kwargs)
        return response

    def post(self, endpoint, data, submission_id=None, **kwargs):
        url = get_url(endpoint, hostname=self.hostname, submission_id=submission_id)
        response = self._request(requests.post, url=url, json=data, **kwargs)
        return response

    def _request(self, func, url, **kwargs):
        response = func(url=url, headers=self.headers, **kwargs)
        assert response.status_code == 200, response.json()
        return response.json()


@auto_authenticate
class SubmitterClient(_ChaiverseHTTPClient):
    def __init__(self,
            developer_key=None,
            hostname=BASE_SUBMITTER_URL):
        super().__init__(developer_key, hostname)


@auto_authenticate
class FeedbackClient(_ChaiverseHTTPClient):
    def __init__(self,
            developer_key=None,
            hostname=BASE_FEEDBACK_URL):
        super().__init__(developer_key, hostname)




__all__ = ["get_leaderboard"]


import numpy as np
import pandas as pd

from chaiverse.lib import binomial_tools
from chaiverse.utils import get_submissions, distribute_to_workers
from chaiverse.metrics.feedback_metrics import FeedbackMetrics
from chaiverse import constants, feedback


def get_leaderboard(
        developer_key=None,
        max_workers=constants.DEFAULT_MAX_WORKERS,
        submission_date_range=None,
        evaluation_date_range=None,
        submission_ids=None,
        fetch_feedback=False,
        ):
    submissions = get_submissions(developer_key, submission_date_range)
    submissions = _filter_submissions_by_submission_ids(submissions, submission_ids) if submission_ids != None else submissions
    submissions = _filter_submissions_by_feedback_count(submissions, constants.PUBLIC_LEADERBOARD_MINIMUM_FEEDBACK_COUNT)
    df = distribute_to_workers(
        get_leaderboard_row,
        submissions.items(),
        developer_key=developer_key,
        evaluation_date_range=evaluation_date_range,
        max_workers=max_workers,
        fetch_feedback=fetch_feedback
    )
    df = pd.DataFrame(df)
    if len(df):
        df = _get_filled_leaderboard(df)
        df.index = np.arange(1, len(df)+1)
    return df


def get_leaderboard_row(submission_item, developer_key=None, evaluation_date_range=None, fetch_feedback=False):
    submission_id, submission_data = submission_item
    submission_feedback_total = submission_data['thumbs_up'] + submission_data['thumbs_down']
    is_updated = feedback.is_submission_updated(submission_id, submission_feedback_total)

    feedback_metrics = {
        'thumbs_up_ratio': binomial_tools.get_ratio(submission_data['thumbs_up'], submission_data['thumbs_down']),
        'thumbs_up_ratio_se': binomial_tools.get_ratio_se(submission_data['thumbs_up'], submission_data['thumbs_down']),
        'total_feedback_count': submission_data['thumbs_up'] + submission_data['thumbs_down'],
    }
    if fetch_feedback:
        feedback_metrics = get_submission_metrics(
            submission_id, 
            developer_key, 
            reload=is_updated, 
            evaluation_date_range=evaluation_date_range
        )
    return {'submission_id': submission_id, **submission_data, **feedback_metrics}


def _get_filled_leaderboard(df):
    # maintain backwards compatibility with model_name field
    _fill_default_value(df, 'model_name', df['submission_id'])
    _fill_default_value(df, 'is_custom_reward', False)
    for col in _get_filled_columns():
        _fill_default_value(df, col, None)
    return df


def _get_filled_columns():
    columns = []
    for competition_type in constants.COMPETITION_TYPE_CONFIGURATION.keys():
        new_columns = constants.COMPETITION_TYPE_CONFIGURATION[competition_type]['output_columns']
        columns.extend(new_columns)
    return columns


def _fill_default_value(df, field, default_value):
    if field not in df:
        df[field] = None
    if default_value is not None:
        df[field] = df[field].fillna(default_value)


def _filter_submissions_by_submission_ids(submissions, submission_ids):
    filtered_submissions = {
        submission_id: data
        for submission_id, data in submissions.items()
        if submission_id in submission_ids
    }
    return filtered_submissions


def _filter_submissions_by_feedback_count(submissions, min_feedback_count):
    submissions = {
        submission_id: submission_data for submission_id, submission_data in submissions.items()
        if submission_data['thumbs_up'] + submission_data['thumbs_down'] >= min_feedback_count
    }
    return submissions


def get_submission_metrics(submission_id, developer_key, reload=True, evaluation_date_range=None):
    feedback_data = feedback.get_feedback(submission_id, developer_key, reload=reload)
    feedback_metrics = FeedbackMetrics(feedback_data.raw_data)
    feedback_metrics.filter_for_date_range(evaluation_date_range)
    feedback_metrics.filter_duplicated_uid()
    metrics = feedback_metrics.calc_metrics()
    return metrics



from datetime import datetime, timezone
import pytz
from typing import Literal


US_PACIFIC = pytz.timezone('US/Pacific')
UTC = pytz.timezone('UTC')


def is_epoch_time_in_date_range(epoch_time, date_range):
    start_epoch_time = _get_date_range_field(date_range, 'start_date') or 0
    end_epoch_time = _get_date_range_field(date_range, 'end_date') or float('inf')
    is_in = start_epoch_time < epoch_time < end_epoch_time
    return is_in


def us_pacific_string(date_string):
    return _create_date_string_in_timezone(date_string, US_PACIFIC)


def utc_string(date_string):
    return _create_date_string_in_timezone(date_string, UTC)


def convert_to_utc_iso_format(date_string=None, default_timezone=US_PACIFIC):
    date_string_in_utc = None
    if date_string is not None:
        date = datetime.fromisoformat(date_string)
        date = default_timezone.localize(date) if not date.tzinfo else date
        date_in_utc = date.astimezone(timezone.utc)
        date_string_in_utc = date_in_utc.isoformat()
    return date_string_in_utc


def _create_date_string_in_timezone(date_string, pytz_timezone):
    date = datetime.fromisoformat(date_string)
    date = pytz_timezone.localize(date)
    return date.isoformat()


def _get_date_range_field(date_range=None, field=Literal['start_date', 'end_date']):
    epoch_time = None
    if date_range and date_range.get(field):
        isoformat_date = date_range.get(field)
        date = datetime.fromisoformat(isoformat_date) 
        assert date.tzinfo
        epoch_time = date.timestamp()
    return epoch_time



__all__ = ["FeedbackMetrics"]


from collections import defaultdict

import numpy as np

from chaiverse.lib import date_tools
from chaiverse.metrics.conversation_metrics import ConversationMetrics


class FeedbackMetrics():
    def __init__(self, feedback_data):
        feedback_dict = feedback_data['feedback']
        feedback_dict = _insert_server_epoch_time(feedback_dict)
        self.feedbacks = list(feedback_dict.values())

    def filter_duplicated_uid(self):
        self.feedbacks = _filter_duplicated_uid_feedbacks(self.feedbacks)

    def filter_for_date_range(self, evaluation_date_range):
        self.feedbacks = [
            feedback for feedback in self.feedbacks
            if date_tools.is_epoch_time_in_date_range(feedback['server_epoch_time'], evaluation_date_range)
        ]

    def calc_metrics(self):
        metrics = {}
        if len(self.convo_metrics) > 0:
            metrics = {
                'mcl': self.mcl,
                'thumbs_up_ratio': self.thumbs_up_ratio,
                'thumbs_up_ratio_se': self.thumbs_up_ratio_se,
                'repetition': self.repetition_score,
                'total_feedback_count': self.total_feedback_count,
            }
        return metrics

    @property
    def convo_metrics(self):
        return [ConversationMetrics(feedback['messages']) for feedback in self.feedbacks]

    @property
    def thumbs_up_ratio(self):
        is_thumbs_up = [feedback['thumbs_up'] for feedback in self.feedbacks]
        thumbs_up = sum(is_thumbs_up)
        thumbs_up_ratio = np.nan if not thumbs_up else thumbs_up / len(is_thumbs_up)
        return thumbs_up_ratio

    @property
    def thumbs_up_ratio_se(self):
        num = self.thumbs_up_ratio * (1 - self.thumbs_up_ratio)
        denom = self.total_feedback_count**0.5
        se = np.nan if self.total_feedback_count < 2 else num / denom
        return se

    @property
    def total_feedback_count(self):
        return len(self.feedbacks)

    @property
    def mcl(self):
        return np.mean([m.mcl for m in self.convo_metrics])

    @property
    def repetition_score(self):
        scores = np.array([m.repetition_score for m in self.convo_metrics])
        is_public = np.array([feedback.get('public', True) for feedback in self.feedbacks])
        return np.nanmean(scores[is_public])


def _insert_server_epoch_time(feedback_dict):
    for feedback_id, feedback in feedback_dict.items():
        feedback['server_epoch_time'] = int(feedback_id.split('_')[-1])
    return feedback_dict


def _filter_duplicated_uid_feedbacks(feedbacks):
    user_feedbacks = defaultdict(list)
    for feedback in feedbacks:
        user_id = feedback["conversation_id"].split("_")[3]
        user_feedbacks[user_id].append(feedback)
    feedbacks = [metrics[0] for _, metrics in user_feedbacks.items()]
    return feedbacks


